import numpy as np
def compute_gcd(X):
    if len(X)==1:
        return X
    elif len(X)==2:
        x,y=X
    else:
        return compute_gcd([compute_gcd(X[0:2]),compute_gcd(X[2:])])
    while(y):
        x, y = y, x % y
    return x

def MWs(table_file):
    mws_dict = {}
    with open(table_file) as f:
        lines = f.readlines()
        for li in lines:
            li_lst = li.split()
            if '#' not in li_lst[0]:
                mws_dict[li_lst[2]] = float(li_lst[0])/1000
    return mws_dict

def txt2dict(txt):
    lines=txt.split('\n')
    keys=lines[0][1:].split()
    vals_arr=[]
    for linei in lines[1:]:
        vals = linei.split()
        if len(vals)==len(keys):
            vals_arr.append([float(vali) for vali in vals])
    vals_arr=np.array(vals_arr)
    res_dict={}
    for k,v in zip(keys,vals_arr.T):
        res_dict[k]=v
    return res_dict



import debyetools.tpropsgui.functions0 as fn
import PySimpleGUI as sg
sg.set_options(element_padding=(0, 0))

from matplotlib import pyplot as plt
plt.rcParams["font.family"] = "serif"
plt.rcParams["mathtext.fontset"] = "dejavuserif"

class lines:
    def __init__(self,values):
        self.keys_list=[]
        for k in range(len(values['from_tab'])):
            self.keys_list.append('l'+str(k))
            setattr(self,'l'+str(k),{'from_tab':values['from_tab'][k], 'x':values['xs'][k], 'y':values['ys'][k], 'label':values['labels'][k]})

    def keys(self):
        return self.keys_list

class tabs:
    def __init__(self, tabs_multilinetxt):
        for tm in tabs_multilinetxt:
            setattr(self, tm, tabs_multilinetxt[tm])
    def keys(self):
        return list(self.__dict__.keys())

class dataplot:
    def __init__(self):
        pass
    def load_tab_data(self,tabs):
        self.tabs = tabs

        values=fn.values_from_tabs(self.tabs)
        self.lines=lines(values)

    def load_lines_settings(self,line_settings):
        self.lines_settings = line_settings
        for k in self.lines.keys():
            try:
                getattr(self.lines,k)['settings'] =line_settings[k]
            except:
                getattr(self.lines,k)['settings'] = {'plot':True,'label':0,'linestyle':'-','color':'gray','marker':'x','markerfacecolor':'gray','markeredgecolor':'cornflowerblue','linewidth':2,'markersize':8}

    def create_window(self,simple=False):
        if simple == True:
            print('simple_layout')
            layout = fn.simple_layout(self.lines)
        else:
            layout = fn.general_layout(self.lines)
        window = sg.Window('Plotting stuff...', layout, finalize=True)
        self.window=window

    def create_popupwindow(self):
        layout = fn.popup_layout(self.tabs)
        window = sg.Window('Edit data...', layout, finalize=True)
        self.popup_window=window

    def update_window(self, fig_settings):
        self.fig_settings=fig_settings
        for k in fig_settings.keys():
            self.window[k].update(fig_settings[k])

    def create_canvas(self,show=False):
        print('create_canvas show',show)

        self.fig = fn.plot_fig(self.fig_settings,self.lines,show=show)
        self.figure_canvas_agg = fn.draw_figure(self.window['--CANVAS2-'].TKCanvas, self.fig)

    def delete_fig_agg(self):
        self.figure_canvas_agg.get_tk_widget().forget()
        plt.close('all')

    def update_canvas(self,show=False):
        print('update_canvas show',show)

        self.delete_fig_agg()
        self.create_canvas(show=show)

    def increment_b_updn(self,event, values, incr, rdval):
        udstr = '_'+event.replace('_',' ').split()[-1]
        if udstr=='_UP':posneg=1
        elif udstr=='_DN':posneg=-1
        keywd = event.replace('--B_','').replace(udstr,'')
        counter = float(values[keywd])
        counter += posneg*incr
        self.window[keywd].update(round(counter,rdval))

    def update_formats(self):
        for k in self.fig_settings.keys():
            self.fig_settings[k]=self.window[k].get()

    def update_lines_settings(self):
        for l in list(self.window.key_dict.keys()):
            if l[:2]=='++':
                prp = l.replace('++',' ').split()
                li = getattr(self.lines,prp[-1])['settings']
                li[prp[0]]=self.window[l].get()

    def copy_multiline2dic(self,add=False):
        for l in list(self.popup_window.key_dict.keys()):
            if l[:2]=='||':
                prp = l.replace('||',' ').split()
                try:
                    getattr(self.tabs,prp[-1])['multiline'] = self.popup_window[l].get()
                except:
                    pass
        if add:
            setattr(self.tabs,'t'+str(int(prp[-1].replace('t',''))+1),{'multiline':'#new$~$Y(X)\n'})

    def update_lines(self):
        keys = self.lines.keys()



import numpy as np
from debyetools.debfunct import D_3, dD_3dx, d2D_3dx2, d3D_3dx3
hbar = 0.1054571800e-33
NAv  = 0.6022140857e24
kB   = 0.138064852e-22
r = 1
class Vibrational:
    """
    Instanceiate the vibrational contribution to the free energy and its derivatives for the calculation of the thermodynamic properties.

    :param float nu: Poisson's ratio.
    :param potential_instance EOS_obj: Equation of state object.
    :param float m: Mass in Kg/mol-at.
    :param intAnharmonicity_instance intanh: Intrinsic anharmonicity object.
    """
    def __init__(self, nu, EOS_obj, m, intanh):
        self.EOS = EOS_obj
        self.kv = (2./3.*((2. + 2.*nu)/(3.-6.*nu))**(3./2.) + 1./3.*((1. + nu)/(3. - 3.*nu))**(3./2.))**(-1./3.)
        self.m = m
        self.intanh = intanh

        self.xDcte = hbar*6**(1/3.)*(np.pi**2*NAv*r)**(1/3.)

    def set_int_anh(self, T, V):
        """
        Calculates intrinsic anharmonicity correction to the Debye temperature and its derivatibes.

        :param float T: Temperature.
        :param float V: Volume.
        """
        self.Anh         = self.intanh.Anh(T,V)
        self.dAnhdT_V    = self.intanh.dAnhdT_V()
        self.dAnhdV_T    = self.intanh.dAnhdV_T(T,V)
        self.d2AnhdVdT   = self.intanh.d2AnhdVdT(T)
        self.d2AnhdV2_T  = self.intanh.d2AnhdV2_T(T,V)
        self.d2AnhdT2_V  = self.intanh.d2AnhdT2_V()
        self.d3AnhdV2dT  = self.intanh.d3AnhdV2dT(T,V)
        self.d3AnhdVdT2  = self.intanh.d3AnhdVdT2(T)
        self.d3AnhdV3_T  = self.intanh.d3AnhdV3_T(T,V)
        self.d4AnhdV4_T  = self.intanh.d4AnhdV4_T(T,V)

    def set_theta(self, T, V, V0_DM, a_DM, b_DM):
        """
        Calculates the Debye Temperature and its derivatives.

        :param float T: Temperature.
        :param float V: Volume.
        """
        kv = self.kv
        m = self.m
        dE0dV_T = self.EOS.dE0dV_T(V)
        d2E0dV2_T = self.EOS.d2E0dV2_T(V)
        d3E0dV3_T = self.EOS.d3E0dV3_T(V)
        d4E0dV4_T = self.EOS.d4E0dV4_T(V)
        d5E0dV5_T = self.EOS.d5E0dV5_T(V)
        d6E0dV6_T = self.EOS.d6E0dV6_T(V)
        P0      = - dE0dV_T
        dP0dV   = - d2E0dV2_T
        d2P0dV2 = - d3E0dV3_T
        d3P0dV3 = - d4E0dV4_T
        d4P0dV4 = - d5E0dV5_T
        d5P0dV5 = - d6E0dV6_T
        vDPrm  = - dP0dV/(r*m)
        vDsqrt = np.sqrt( vDPrm)
        vD      = kv*V*vDsqrt
        dvDdV   = kv*vDsqrt-kv*V*d2P0dV2/(2*vDsqrt*r*m)
        d2vDdV2 = -kv*d2P0dV2/(vDsqrt*r*m)-kv*V*d2P0dV2**2/(4*(vDPrm)**(3/2)*r**2*m**2)-kv*V*d3P0dV3/(2*vDsqrt*r*m)
        d3vDdV3 = -3*kv*d2P0dV2**2/(4*(vDPrm)**(3/2)*r**2*m**2)-3*kv*d3P0dV3/(2*vDsqrt*r*m)-3*kv*V*d2P0dV2**3/(8*(vDPrm)**(5/2)*r**3*m**3)-3*kv*V*d2P0dV2*d3P0dV3/(4*(vDPrm)**(3/2)*r**2*m**2)-kv*V*d4P0dV4/(2*vDsqrt*r*m)
        d4vDdV4 = -3*kv*(d2P0dV2)**3/(2*(-dP0dV/(r*m))**(5/2)*r**3*m**3)-3*kv*(d2P0dV2)*(d3P0dV3)/((-dP0dV/(r*m))**(3/2)*r**2*m**2)-2*kv*(d4P0dV4)/(np.sqrt(-dP0dV/(r*m))*r*m)-15*kv*V*(d2P0dV2)**4/(16*(-dP0dV/(r*m))**(7/2)*r**4*m**4)-9*kv*V*(d2P0dV2)**2*(d3P0dV3)/(4*(-dP0dV/(r*m))**(5/2)*r**3*m**3)-3*kv*V*(d3P0dV3)**2/(4*(-dP0dV/(r*m))**(3/2)*r**2*m**2)-kv*V*(d2P0dV2)*(d4P0dV4)/((-dP0dV/(r*m))**(3/2)*r**2*m**2)-kv*V*(d5P0dV5)/(2*np.sqrt(-dP0dV/(r*m))*r*m)
        xD      = self.xDcte*(1/V)**(1/3.)/kB
        dxDdV   = - self.xDcte/(3*kB*V**(4/3.))
        d2xDdV2 = 4*self.xDcte/(9*kB*V**(7/3.))
        d3xDdV3 = -28*self.xDcte/(27*V**(10/3)*kB)
        d4xDdV4 = 280*self.xDcte/(81*V**(13/3)*kB)

        B0_DM = V0_DM*self.EOS.d2E0dV2_T(V0_DM)
        DM = 1#(V*d2E0dV2_T/B0_DM)**b_DM/(V/V0_DM)**a_DM
        dDMdV = 0#V0_DM**a_DM*(-d2E0dV2_T**b_DM*(-b_DM+a_DM)*V**(b_DM-1-a_DM)+V**(b_DM-a_DM)*d2E0dV2_T**(b_DM-1)*(d3E0dV3_T)*b_DM)*B0_DM**(-b_DM)
        d2DMdV2 = 0#V0_DM**a_DM*B0_DM**(-b_DM)*(-2*(d3E0dV3_T)*b_DM*d2E0dV2_T**(b_DM-1)*(-b_DM+a_DM)*V**(b_DM-1-a_DM)+d2E0dV2_T**b_DM*(-b_DM+1+a_DM)*(-b_DM+a_DM)*V**(b_DM-2-a_DM)+V**(b_DM-a_DM)*(d2E0dV2_T**(b_DM-1)*(d4E0dV4_T)+(d3E0dV3_T)**2*d2E0dV2_T**(b_DM-2)*(b_DM-1))*b_DM)
        d3DMdV3 = 0#-V0_DM**a_DM*B0_DM**(-b_DM)*(d2E0dV2_T**b_DM*(-b_DM+2+a_DM)*(-b_DM+1+a_DM)*(-b_DM+a_DM)*V**(b_DM-a_DM-3)-3*b_DM*(-(d2E0dV2_T**(b_DM-1)*(d4E0dV4_T)+(d3E0dV3_T)**2*d2E0dV2_T**(b_DM-2)*(b_DM-1))*(-b_DM+a_DM)*V**(b_DM-1-a_DM)+(d3E0dV3_T)*d2E0dV2_T**(b_DM-1)*(-b_DM+1+a_DM)*(-b_DM+a_DM)*V**(b_DM-2-a_DM)+(1/3)*V**(b_DM-a_DM)*(d2E0dV2_T**(b_DM-1)*(d5E0dV5_T)+(d3E0dV3_T)*(3*d2E0dV2_T**(b_DM-2)*(d4E0dV4_T)+(d3E0dV3_T)**2*d2E0dV2_T**(b_DM-3)*(b_DM-2))*(b_DM-1))))
        d4DMdV4 = 0#(d2E0dV2_T**b_DM*(-b_DM+a_DM)*(-b_DM+a_DM+3)*(-b_DM+2+a_DM)*(-b_DM+1+a_DM)*V**(b_DM-4-a_DM)+6*b_DM*((d2E0dV2_T**(b_DM-1)*(d4E0dV4_T)+(d3E0dV3_T)**2*d2E0dV2_T**(b_DM-2)*(b_DM-1))*(-b_DM+a_DM)*(-b_DM+1+a_DM)*V**(b_DM-2-a_DM)-(1/3)*(2*(d2E0dV2_T**(b_DM-1)*(d5E0dV5_T)+(d3E0dV3_T)*(3*d2E0dV2_T**(b_DM-2)*(d4E0dV4_T)+(d3E0dV3_T)**2*d2E0dV2_T**(b_DM-3)*(b_DM-2))*(b_DM-1)))*(-b_DM+a_DM)*V**(b_DM-1-a_DM)-2*(d3E0dV3_T)*d2E0dV2_T**(b_DM-1)*(-b_DM+2+a_DM)*(-b_DM+1+a_DM)*(-b_DM+a_DM)*V**(b_DM-a_DM-3)*(1/3)+(1/6)*(d2E0dV2_T**(b_DM-1)*(d6E0dV6_T)+(4*(d3E0dV3_T)*d2E0dV2_T**(b_DM-2)*(d5E0dV5_T)+3*d2E0dV2_T**(b_DM-2)*(d4E0dV4_T)**2+(6*d2E0dV2_T**(b_DM-3)*(d4E0dV4_T)+(d3E0dV3_T)**2*d2E0dV2_T**(b_DM-4)*(b_DM-3))*(b_DM-2)*(d3E0dV3_T)**2)*(b_DM-1))*V**(b_DM-a_DM)))*V0_DM**a_DM*B0_DM**(-b_DM)

        self.tD        = xD*vD*DM*self.Anh

        B0_DM = V0_DM*self.EOS.d2E0dV2_T(V0_DM)
        DM = (V*d2E0dV2_T/B0_DM)**b_DM/(V/V0_DM)**a_DM
        Anh         = self.intanh.Anh(T,V)
        xDcte = self.xDcte
        xD      = xDcte*(1/V)**(1/3.)/kB
        vDPrm  = - dP0dV/(r*m)
        vDsqrt = np.sqrt( vDPrm)
        vD      = kv*V*vDsqrt
        self.tD        = xD*vD*Anh*DM


        self.dtDdV_T   = dxDdV*vD*self.Anh*DM+xD*dvDdV*self.Anh*DM+xD*vD*self.dAnhdV_T*DM+xD*vD*self.Anh*dDMdV
        self.dtDdT_V   = xD*vD*DM*self.dAnhdT_V
        self.d2tDdV2_T = xD*vD*self.d2AnhdV2_T*DM+xD*vD*self.Anh*d2DMdV2+xD*d2vDdV2*self.Anh*DM+d2xDdV2*vD*self.Anh*DM+(2*vD*xD*dDMdV+2*DM*(dxDdV*vD+xD*dvDdV))*self.dAnhdV_T+(2*((dxDdV*vD+xD*dvDdV)*dDMdV+DM*dvDdV*dxDdV))*self.Anh
        self.d2tDdT2_V = xD*vD*DM*self.d2AnhdT2_V
        self.d2tDdVdT  = xD*vD*self.d2AnhdVdT*DM+(vD*xD*dDMdV+DM*(dxDdV*vD+xD*dvDdV))*self.dAnhdT_V
        self.d3tDdV3_T = xD*vD*self.d3AnhdV3_T*DM+xD*vD*self.Anh*d3DMdV3+xD*d3vDdV3*self.Anh*DM+d3xDdV3*vD*self.Anh*DM+(3*vD*xD*dDMdV+3*DM*(dxDdV*vD+xD*dvDdV))*self.d2AnhdV2_T+(3*self.dAnhdV_T*vD*xD+3*self.Anh*(dxDdV*vD+xD*dvDdV))*d2DMdV2+(3*self.dAnhdV_T*DM*xD+3*self.Anh*(DM*dxDdV+xD*dDMdV))*d2vDdV2+(3*self.dAnhdV_T*DM*vD+3*self.Anh*(DM*dvDdV+vD*dDMdV))*d2xDdV2+((6*dxDdV*vD+6*xD*dvDdV)*dDMdV+6*DM*dvDdV*dxDdV)*self.dAnhdV_T+6*dxDdV*dvDdV*self.Anh*dDMdV
        self.d3tDdV2dT = xD*vD*self.d3AnhdV2dT*DM+(2*vD*xD*dDMdV+2*DM*(dxDdV*vD+xD*dvDdV))*self.d2AnhdVdT+(d2DMdV2*vD*xD+d2vDdV2*DM*xD+d2xDdV2*DM*vD+(2*dxDdV*vD+2*xD*dvDdV)*dDMdV+2*DM*dvDdV*dxDdV)*self.dAnhdT_V
        self.d3tDdVdT2 = xD*vD*self.d3AnhdVdT2*DM+(vD*xD*dDMdV+DM*(dxDdV*vD+xD*dvDdV))*self.d2AnhdT2_V
        self.d4tDdV4_T = xD*vD*self.d4AnhdV4_T*DM+xD*vD*self.Anh*d4DMdV4+xD*d4vDdV4*self.Anh*DM+d4xDdV4*vD*self.Anh*DM+(4*vD*xD*dDMdV+4*DM*(dxDdV*vD+xD*dvDdV))*self.d3AnhdV3_T+(4*self.dAnhdV_T*vD*xD+4*self.Anh*(dxDdV*vD+xD*dvDdV))*d3DMdV3+(4*self.dAnhdV_T*DM*xD+4*self.Anh*(DM*dxDdV+xD*dDMdV))*d3vDdV3+(4*self.dAnhdV_T*DM*vD+4*self.Anh*(DM*dvDdV+vD*dDMdV))*d3xDdV3+(6*d2DMdV2*vD*xD+6*d2vDdV2*DM*xD+6*d2xDdV2*DM*vD+(12*dxDdV*vD+12*xD*dvDdV)*dDMdV+12*DM*dvDdV*dxDdV)*self.d2AnhdV2_T+(6*d2vDdV2*xD*self.Anh+ 6*d2xDdV2*vD*self.Anh + (12*dxDdV*vD+12*xD*dvDdV)*self.dAnhdV_T + 12*self.Anh*dvDdV*dxDdV)*d2DMdV2 + (6*d2xDdV2*DM*self.Anh+(12*DM*dxDdV+12*xD*dDMdV)*self.dAnhdV_T+12*self.Anh*dDMdV*dxDdV)*d2vDdV2+((12*DM*dvDdV+12*vD*dDMdV)*self.dAnhdV_T+12*self.Anh*dDMdV*dvDdV)*d2xDdV2+24*dxDdV*dvDdV*self.dAnhdV_T*dDMdV

    def E(self, T, V):
        x = self.tD/T
        D3 = D_3(x)

        return -(3*(self.dtDdT_V*T-self.tD))*r*NAv*kB*(T*D3+3*self.tD*(1/8))/self.tD

    def S(self, T, V):
        x = self.tD/T
        D3 = D_3(x)

        return -3*kB*(np.log(1-np.exp(-self.tD/T))*self.tD+(self.dtDdT_V*T-4*self.tD*(1/3))*D3+3*self.dtDdT_V*self.tD*(1/8))*r*NAv/self.tD
    def Fmin(self, T, V, V0_DM, a_DM, b_DM):
        """
        Vibration Helmholtz free energy.

        :param float T: Temperature.
        :param float V: Volume.
        """
        d2E0dV2_T_0 = self.EOS.d2E0dV2_T(V0_DM)
        d2E0dV2_T = self.EOS.d2E0dV2_T(V)
        dE0dV_T_0 = self.EOS.dE0dV_T(V0_DM)
        if type(V) == np.ndarray:
            if d2E0dV2_T_0<0:return 1
        kv = self.kv
        m = self.m

        dP0dV_0   = - d2E0dV2_T_0

        B0_DM = V0_DM*self.EOS.d2E0dV2_T(V0_DM)
        DM = (V*d2E0dV2_T/B0_DM)**b_DM/(V/V0_DM)**a_DM
        Anh         = self.intanh.Anh(T,V)
        xDcte = self.xDcte
        xD0      = xDcte*(1/V0_DM)**(1/3.)/kB
        vDPrm0  = - dP0dV_0/(r*m)
        vDsqrt0 = np.sqrt( vDPrm0)
        vD0      = kv*V0_DM*vDsqrt0
        tD0        = xD0*vD0*Anh*DM

        x = tD0/T
        D3 = D_3(x)
        # print(tD/T, tD, T)
        if type(V) is not np.ndarray:
            if x<0.07:return 1
        return 3*r*NAv*kB*(tD0*3/8 + T*np.log(1-np.exp(-x))  - D3*T/3)

    def F(self, T, V, V0_DM, a_DM, b_DM):
        """
        Vibration Helmholtz free energy.

        :param float T: Temperature.
        :param float V: Volume.
        """
        d2E0dV2_T = self.EOS.d2E0dV2_T(V)
        dE0dV_T = self.EOS.dE0dV_T(V)
        if type(V) == np.ndarray:
            if min(d2E0dV2_T)<0:return 1
        kv = self.kv
        m = self.m

        P0 = - dE0dV_T
        dP0dV   = - d2E0dV2_T

        B0_DM = V0_DM*self.EOS.d2E0dV2_T(V0_DM)
        DM = (V*d2E0dV2_T/B0_DM)**b_DM/(V/V0_DM)**a_DM
        Anh         = self.intanh.Anh(T,V)
        xDcte = self.xDcte
        xD      = xDcte*(1/V)**(1/3.)/kB
        vDPrm  = - dP0dV/(r*m)
        vDsqrt = np.sqrt( vDPrm)
        vD      = kv*V*vDsqrt
        tD        = xD*vD*Anh*DM

        x = tD/T
        D3 = D_3(x)
        # print(tD/T, tD, T)
        if type(V) is not np.ndarray:
            if x<0.07:return 1
        return 3*r*NAv*kB*(tD*3/8 + T*np.log(1-np.exp(-x))  - D3*T/3)

    # def dSdV_T(self, T, V):
    #     _dSvibdV_T =  -3*kB*(_dtDdV_T*np.exp(-_x)*_tD/(_T*(1-np.exp(-_x)))+np.log(1-np.exp(-_x))*_dtDdV_T+(_d2tDdVdT*_T-4*_dtDdV_T*(1/3))*_D3+(_dtDdT_V*_T-4*_tD*(1/3))*_dD3dx*_dtDdV_T/_T+3*_d2tDdVdT*_tD*(1/8)+3*_dtDdT_V*_dtDdV_T*(1/8))*r*NAv/_tD+3*kB*(np.log(1-np.exp(-_x))*_tD+(_dtDdT_V*_T-4*_tD*(1/3))*_D3+3*_dtDdT_V*_tD*(1/8))*r*NAv*_dtDdV_T/_tD**2
    #
    # def dFdV_T(self,T,V):
    #     d2E0dV2_T = self.EOS.d2E0dV2_T
    #     if type(V) == np.ndarray:
    #         if min(d2E0dV2_T)<0:return 1
    #     kv = self.kv
    #     m = self.m
    #
    #     d3E0dV3_T = self.EOS.d3E0dV3_T(V)
    #
    #     dP0dV   = - d2E0dV2_T
    #     d2P0dV2 = - d3E0dV3_T
    #
    #     Anh      = self.intanh.Anh(T,V)
    #     dAnhdV_T = self.intanh.dAnhdV_T(T,V)
    #     xDcte = self.xDcte
    #     xD      = xDcte*(1/V)**(1/3.)/kB
    #     dxDdV   = - self.xDcte/(3*kB*V**(4/3.))
    #     vDPrm  = - dP0dV/(r*m)
    #     vDsqrt = np.sqrt( vDPrm)
    #     vD      = kv*V*vDsqrt
    #     dvDdV   = kv*vDsqrt-kv*V*d2P0dV2/(2*vDsqrt*r*m)
    #     tD        = xD*vD*Anh
    #     dtDdV_T = dxDdV*vD*Anh + xD*dvDdV*Anh + xD*vD*dAnhdV_T
    #
    #     x = tD/T
    #     D3 = D_3(x)
    #     dD3dx   = dD_3dx(x,D3)
    #
    #     return 3*r*NAv*kB*(3*dtDdV_T*(1/8)+dtDdV_T*np.exp(-x)/(1-np.exp(-x))-(1/3)*dD3dx*dtDdV_T)

    def d2FdT2_V(self,T,V):
        """
        (d2F(T,V)/dT2)_V

        :param float T: Temperature.
        :param float V: Volume.
        """
        x = self.tD/T
        ex = np.exp(x)
        D3      = D_3(x)
        return 3*r*NAv*((ex-1)*T*(T**2*self.d2tDdT2_V*self.tD-4*(self.dtDdT_V*T-self.tD)**2)*D3+3*self.tD*(self.d2tDdT2_V*self.tD*ex*T**2-T**2*self.d2tDdT2_V*self.tD+8*(self.dtDdT_V*T-self.tD)**2)*(1/8))*kB/(self.tD**2*(ex-1)*T**2)

    def d2FdV2_T(self,T,V):
        """
        (d2F(T,V)/dV2)_T

        :param float T: Temperature.
        :param float V: Volume.
        """
        x = self.tD/T
        D3      = D_3(x)
        dD3dx   = dD_3dx(x,D3)
        return 3*r*NAv*kB*(8*self.dtDdV_T**2*self.tD*dD3dx-8*self.dtDdV_T**2*D3*T+8*self.d2tDdV2_T*D3*self.tD*T+3*self.d2tDdV2_T*self.tD**2)/(8*self.tD**2)

    def d3FdV3_T(self,T,V):
        """
        (d3F(T,V)/dV3)_T

        :param float T: Temperature.
        :param float V: Volume.
        """
        x = self.tD/T
        D3      = D_3(x)
        dD3dx   = dD_3dx(x,D3)
        d2D3dx2 = d2D_3dx2(x,D3,dD3dx)
        return 3*r*(T**2*(self.d3tDdV3_T*self.tD**2-3*self.dtDdV_T*self.d2tDdV2_T*self.tD+2*self.dtDdV_T**3)*D3+(3*self.dtDdV_T*self.d2tDdV2_T*T*self.tD**2-2*self.dtDdV_T**3*T*self.tD)*dD3dx+d2D3dx2*self.dtDdV_T**3*self.tD**2+3*self.d3tDdV3_T*self.tD**3*T*(1/8))*kB*NAv/(T*self.tD**3)

    def d4FdV4_T(self,T,V):
        """
        (d4F(T,V)/dV4)_T

        :param float T: Temperature.
        :param float V: Volume.
        """
        x = self.tD/T
        D3      = D_3(x)
        dD3dx   = dD_3dx(x,D3)
        d2D3dx2 = d2D_3dx2(x,D3,dD3dx)
        d3D3dx3 = d3D_3dx3(x,D3,dD3dx,d2D3dx2)

        return -12*r*(T**3*(-(1/4)*self.d4tDdV4_T*self.tD**3+(self.d3tDdV3_T*self.dtDdV_T+3*self.d2tDdV2_T**2*(1/4))*self.tD**2-3*self.dtDdV_T**2*self.d2tDdV2_T*self.tD+3*self.dtDdV_T**4*(1/2))*D3-self.tD*(T**2*((self.d3tDdV3_T*self.dtDdV_T+3*self.d2tDdV2_T**2*(1/4))*self.tD**2-3*self.dtDdV_T**2*self.d2tDdV2_T*self.tD+3*self.dtDdV_T**4*(1/2))*dD3dx+(1/32)*(3*((16*self.dtDdV_T**2*self.d2tDdV2_T*self.tD*T-8*self.dtDdV_T**4*T)*d2D3dx2+self.tD*(8*d3D3dx3*self.dtDdV_T**4*(1/3)+self.d4tDdV4_T*self.tD*T**2)))*self.tD))*kB*NAv/(T**2*self.tD**4)

    def d2FdVdT(self,T,V):
        """
        (d2F(T,V)/dVdT)

        :param float T: Temperature.
        :param float V: Volume.
        """
        x = self.tD/T
        D3      = D_3(x)
        dD3dx   = dD_3dx(x,D3)
        return 3*r*(dD3dx*(self.dtDdT_V/T-self.tD/T**2)*T+D3+3*self.dtDdT_V*(1/8))*kB*self.dtDdV_T*NAv/self.tD+3*r*(D3*T+3*self.tD*(1/8))*kB*self.d2tDdVdT*NAv/self.tD-3*r*(D3*T+3*self.tD*(1/8))*kB*self.dtDdV_T*NAv*self.dtDdT_V/self.tD**2

    def d3FdV2dT(self,T,V):
        """
        (d3F(T,V)/dV2dT)

        :param float T: Temperature.
        :param float V: Volume.
        """
        x = self.tD/T
        D3      = D_3(x)
        dD3dx   = dD_3dx(x,D3)
        d2D3dx2 = d2D_3dx2(x,D3,dD3dx)

        return -3*r*(T**2*((-T*self.d3tDdV2dT-self.d2tDdV2_T)*self.tD**2+(T*self.d2tDdV2_T*self.dtDdT_V+2*T*self.dtDdV_T*self.d2tDdVdT+self.dtDdV_T**2)*self.tD-2*T*self.dtDdT_V*self.dtDdV_T**2)*D3-self.tD*((-self.d2tDdV2_T*self.tD**2+(T*self.d2tDdV2_T*self.dtDdT_V+2*T*self.dtDdV_T*self.d2tDdVdT+self.dtDdV_T**2)*self.tD-2*T*self.dtDdT_V*self.dtDdV_T**2)*T*dD3dx+3*self.tD*(8*self.dtDdV_T**2*(T*self.dtDdT_V-self.tD)*d2D3dx2*(1/3)+self.tD*self.d3tDdV2dT*T**2)*(1/8)))*kB*NAv/(T**2*self.tD**3)

    def d3FdVdT2(self,T,V):
        """
        (d3F(T,V)/dVdT2)

        :param float T: Temperature.
        :param float V: Volume.
        """
        x = self.tD/T
        D3      = D_3(x)
        dD3dx   = dD_3dx(x,D3)
        d2D3dx2 = d2D_3dx2(x,D3,dD3dx)

        return -(6*(T**3*((-(1/2)*self.d3tDdVdT2*T-self.d2tDdVdT)*self.tD**2+(((1/2)*self.d2tDdT2_V*T+self.dtDdT_V)*self.dtDdV_T+self.dtDdT_V*self.d2tDdVdT*T)*self.tD-self.dtDdT_V**2*self.dtDdV_T*T)*D3-(T**2*(-self.tD**2*self.d2tDdVdT+(((1/2)*self.d2tDdT2_V*T+self.dtDdT_V)*self.dtDdV_T+self.dtDdT_V*self.d2tDdVdT*T)*self.tD-self.dtDdT_V**2*self.dtDdV_T*T)*dD3dx+(1/16)*(3*(8*self.dtDdV_T*(T*self.dtDdT_V-self.tD)**2*d2D3dx2*(1/3)+self.tD*self.d3tDdVdT2*T**3))*self.tD)*self.tD))*r*kB*NAv/(T**3*self.tD**3)



import numpy as np
from scipy.optimize import leastsq

NAv  = 0.6022140857e24
kB   = 0.138064852e-22

class Electronic:
    """
    Implementation of the electronic contribution to the free energy.

    :param float params: N(Ef)(V) function parameters.
    """
    def __init__(self,*params):
        self.pel = params
        self.r=1
        for n,q in enumerate(params):
            setattr(self,'q'+str(n),q)

    def NfV(self,V):
        """
        N(Ef)(V)

        :param float V: Volume.
        """
        return self.q0*V**0 + self.q1*V**1 + self.q2*V**2 + self.q3*V**3
    def dNfVdV_T(self,V):
        return 3*V**2*self.q3+2*V*self.q2+self.q1
    def d2NfVdV2_T(self,V):
        return 6*V*self.q3+2*self.q2
    def d3NfVdV3_T(self,V):
        return 6*self.q3
    def d4NfVdV4_T(self,V):
        return 0

    def E(self,T,V):
        return (1/6)*np.pi**2*NAv*self.r*kB**2*T**2*self.NfV(V)/(0.160218e-18)
    def S(self,T,V):
        return (2/6)*np.pi**2*NAv*self.r*kB**2*T*self.NfV(V)/(0.160218e-18)
    def F(self,T,V):
        """
        Electronic contribution to the free energy.

        :param float T: Temperature.
        :param float V: Volume.
        """
        return - (1/6)*np.pi**2*NAv*self.r*kB**2*T**2*self.NfV(V)/(0.160218e-18)
    def dFdV_T(self, T, V):
        return - (1/6)*np.pi**2*NAv*self.r*kB**2*T**2*self.dNfVdV_T(V)/(0.160218e-18)
    def d2FdT2_V(self,T,V):
        return - 2*np.pi**2*NAv*self.r*kB**2*self.NfV(V)*(1/6)/(0.160218e-18)
    def d2FdV2_T(self,T,V):
        return - (1/6)*np.pi**2*NAv*self.r*kB**2*T**2*self.d2NfVdV2_T(V)/(0.160218e-18)
    def d3FdV3_T(self,T,V):
        return - (1/6)*np.pi**2*NAv*self.r*kB**2*T**2*self.d3NfVdV3_T(V)/(0.160218e-18)
    def d4FdV4_T(self,T,V):
        return - (1/6)*np.pi**2*NAv*self.r*kB**2*T**2*self.d4NfVdV4_T(V)/(0.160218e-18)

    def d2FdVdT(self,T,V):
        return -2*np.pi**2*NAv*self.r*kB**2*T*self.dNfVdV_T(V)*(1/6)/(0.160218e-18)
    def d3FdV2dT(self,T,V):
        return -2*np.pi**2*NAv*self.r*kB**2*T*self.d2NfVdV2_T(V)*(1/6)/(0.160218e-18)
    def d3FdVdT2(self,T,V):
        return -2*np.pi**2*NAv*self.r*kB**2*self.dNfVdV_T(V)*(1/6)/(0.160218e-18)

def fit_electronic(Vs, p_el,E,N,Ef):
    """
    Fitting procedure for the N(Ef)(V) function.

    :param list_of_floats Vs: Volumes.
    :param list_of_floats p_el: Initial parameters.
    :param list_of_lists_of_floats E: Matrix with energies at each level for each volume.
    :param list_of_lists_of_floats N: Matrix with densities of state at each level for each volume.
    :param list_of_floats Ef: Fermi levels as function of temperature.

    :retun list_of_floats: optimized parameters.
    """

    V = np.array(Vs)
    ix_V0=10
    EfV0 = float(Ef[ix_V0])
    ixs=[i for i,x in enumerate(E[ix_V0]) if x>=Ef[ix_V0]]
    E1 = float(E[ix_V0][ixs[0]-1])
    E2 = float(E[ix_V0][ixs[0]])
    N1 = float(N[ix_V0][ixs[0]-1])
    N2 = float(N[ix_V0][ixs[0]])
    NfV0 = (EfV0 - E1)*(N2 - N1)/(E2 - E1) + N1
    NfV = np.array([NfV0*np.sqrt(ef/EfV0) for ef in Ef][8:-1])
    P2 = leastsq(NfV2m, p_el, args=(V[8:-1], NfV),maxfev=1000)
    P2 = P2[0]

    return P2

def NfV_poly_fun(V, _A, _B, _C, _D):
    #A, B, C = A*1e-1, B*(-1e4), C*1e-9
    return _A + _B*V + _C*V**2 + _D*V**3

def NfV2m(P, Vdata, NfVdata):
    NfVcalc = [NfV_poly_fun(Vi, P[0], P[1], P[2], P[3]) for Vi in Vdata]
    # print('NfVcalc', NfVcalc, 'NfVdata', NfVdata)
    return NfVcalc-NfVdata



import numpy as np
import itertools as it
import debyetools.aux_functions as afn



def neighbor_list(size, cutoff, center, basis_vectors, primitive_cell):
    """ calculate a list i, j, dij where i and j are a pair of atoms of
    indexes i and j, respectively, and dij is the distance between them.

    :param array size: Number of times we are replicating the primitive cel
    :param int cutoff: Number of times we are replicating the supercell up to which we are considering for the calculation (uses pbc)
    :param array center: The position in space where the system of reference is
    :param array basis_vectors: atoms position within a single primitive cell
    :param array primitive_cell: the primitive cell
    :return: D, I , J
    """

    basis_vectors = np.dot(basis_vectors,primitive_cell)
    size_g = size + 2*cutoff  #change cutoff for some other word and add a cutoff distance.
                              # new_co= int(min(cutoff,....))

    cell_coords_centered = afn.generate_cells_coordinates(size, primitive_cell, center)
    cell_coords_centered_g = afn.generate_cells_coordinates(size_g, primitive_cell, center-cutoff)

    XCs = []
    Is = []
    Js = []

    ixs = np.arange(len(basis_vectors))
    jxs = np.arange(len(basis_vectors))

    for icell, cell_coords_i in enumerate(cell_coords_centered):
        ix_neighbor = np.where(np.all(abs(cell_coords_centered_g-cell_coords_i)<=cutoff, axis=1))[0]
        Xs = np.array([cell_coords_i,]*len(basis_vectors))+basis_vectors

        for cell_coords_i_g in cell_coords_centered_g[ix_neighbor]:
            Xsg = np.array([cell_coords_i_g,]*len(basis_vectors))+basis_vectors
            for x, xg in it.product(Xs, Xsg):
                XCs.append((x-xg)**2)
            for i, j in it.product(ixs, jxs):
                Is.append(i)
                Js.append(j)

    XCs = np.array(XCs)
    Is = np.array(Is)
    Js = np.array(Js)
    CX = np.sum(XCs/cutoff**2, axis=1)

    ix = np.where(np.all([CX<=1, CX>0], axis=0))[0]
    distances = np.sqrt(np.sum(XCs[ix], axis=1))

    return distances, Is[ix], Js[ix]

def pair_analysis(atom_types, size, cutoff, center, basis_vectors, primitive_cell):
    """
    run a pair analysis of a crystal structure of almost any type of symmetry.

    :param list_of_str atom_types: the types of each atom in the primitive cell in the same order as the basis vectors.
    :param array size: Number of times we are replicating the primitive cel
    :param int cutoff: Number of times we are replicating the supercell up to which we are considering for the calculation (uses pbc)
    :param array center: The position in space where the system of reference is
    :param array basis_vectors: atoms position within a single primitive cell
    :param array primitive_cell: the primitive cell
    :return:  pair distance, pair number, pair types
    """
    cutoff = np.array([int(cutoff), int(cutoff), int(cutoff)]) ##<--- fix_here
    dAxBy, iAxBy, jAxBy  = neighbor_list(size, cutoff, center, basis_vectors, primitive_cell)

    nat = np.prod(size)*len(basis_vectors)

    combs_types,types_all = afn.c_types(atom_types)

    dAxBy = np.array([float('%0.10e'%(d)) for d in dAxBy])
    bins_dAxBy =list(set([li for li in list(set(np.append(dAxBy, [max(cutoff)])))]))
    bins_dAxBy.sort()
    distances = bins_dAxBy[:-1]

    ptlst = []
    for i,j,d in zip(iAxBy,jAxBy,dAxBy):
        for ii in range(len(combs_types)):
            if (types_all[i]+'-'+types_all[j] == combs_types[ii]) or (types_all[j]+'-'+types_all[i] == combs_types[ii]):
                pairtype = ii
        ptlst.append(pairtype)

    ptlst=np.array(ptlst)

    ds = ['' for x in range(len(combs_types))]
    for i in range(len(combs_types)):
        ds[i] = np.array([d for d in dAxBy[np.where(ptlst==i)[0]]])
    hs = ['' for x in range(len(combs_types))]
    bs = ['' for x in range(len(combs_types))]
    for i in range(len(combs_types)):
        hs[i], bs[i] = np.histogram(ds[i], bins=bins_dAxBy)
    tot_num_bonds_per_molecule = np.array(hs).T
    num_bonds_per_formula = tot_num_bonds_per_molecule/nat
    return np.array(distances), num_bonds_per_formula, combs_types
